@@ -10,6 +10,7 @@ use Log::Log4perl;
 use Log::Log4perl::Level;
 use Log::Log4perl::Layout;
 use Log::Log4perl::Appender;
+use Log::Log4perl::Appender::String;
 use Log::Log4perl::Filter;
 use Carp;
 
@@ -24,34 +25,52 @@ our $LOGGERS_BY_NAME = {};
 our %APPENDER_BY_NAME = ();
 our $INITIALIZED = 0;
 our $NON_INIT_WARNED;
+our $DIE_DEBUG = 0;
+our $DIE_DEBUG_BUFFER = "";
+    # Define the default appender that's used for formatting
+    # warn/die/croak etc. messages.
+our $STRING_APP_NAME = "_l4p_warn";
+our $STRING_APP      = Log::Log4perl::Appender->new(
+                          "Log::Log4perl::Appender::String",
+                          name => $STRING_APP_NAME);
+$STRING_APP->layout(Log::Log4perl::Layout::PatternLayout->new("%m"));
+our $STRING_APP_CODEREF = generate_coderef([[$STRING_APP_NAME, $STRING_APP]]);
 
 __PACKAGE__->reset();
 
+###########################################
+sub warning_render {
+###########################################
+    my($logger, @message) = @_;
+
+    $STRING_APP->string("");
+    $STRING_APP_CODEREF->($logger, 
+                          @message, 
+                          Log::Log4perl::Level::to_level($ALL));
+    return $STRING_APP->string();
+}
+
 ##################################################
 sub cleanup {
 ##################################################
     # warn "Logger cleanup";
 
+    # Nuke all convenience loggers to avoid them causing cleanup to 
+    # be delayed until global destruction. Problem is that something like
+    #     *{"DEBUG"} = sub { $logger->debug };
+    # ties up a reference to $logger until global destruction, so we 
+    # need to clean up all :easy shortcuts, hence freeing the last
+    # logger references, to then rely on the garbage collector for cleaning
+    # up the loggers.
+    Log::Log4perl->easy_closure_global_cleanup();
+
     # Delete all loggers
-    foreach my $loggername (keys %$LOGGERS_BY_NAME){
-        # warn "Logger delete: $loggername";
-        $LOGGERS_BY_NAME->{$loggername}->DESTROY();
-        delete $LOGGERS_BY_NAME->{$loggername};
-    }
+    $LOGGERS_BY_NAME = {};
 
     # Delete the root logger
     undef $ROOT_LOGGER;
 
     # Delete all appenders
-    foreach my $appendername (keys %APPENDER_BY_NAME){
-        if (exists $APPENDER_BY_NAME{$appendername} &&
-            exists $APPENDER_BY_NAME{$appendername}->{appender}) {
-                # Destroy L4p::Appender
-            $APPENDER_BY_NAME{$appendername}->DESTROY();
-            delete $APPENDER_BY_NAME{$appendername}->{appender};
-        }
-        delete $APPENDER_BY_NAME{$appendername};
-    }
     %APPENDER_BY_NAME   = ();
 
     undef $INITIALIZED;
@@ -60,28 +79,18 @@ sub cleanup {
 ##################################################
 sub DESTROY {
 ##################################################
-    warn "Destroying logger $_[0]" if $Log::Log4perl::CHATTY_DESTROY_METHODS;
-
-    for(keys %{$_[0]}) {
-        delete $_[0]->{$_};
-    }
+    CORE::warn "Destroying logger $_[0] ($_[0]->{category})" 
+            if $Log::Log4perl::CHATTY_DESTROY_METHODS;
 }
 
 ##################################################
 sub reset {
 ##################################################
-    $ROOT_LOGGER        = __PACKAGE__->_new("", $DEBUG);
+    $ROOT_LOGGER        = __PACKAGE__->_new("", $OFF);
 #    $LOGGERS_BY_NAME    = {};  #leave this alone, it's used by 
                                 #reset_all_output_methods when 
                                 #the config changes
 
-
-    #we've got a circular reference thing going on somewhere
-    foreach my $appendername (keys %APPENDER_BY_NAME){
-        delete $APPENDER_BY_NAME{$appendername}->{appender} 
-                if (exists $APPENDER_BY_NAME{$appendername} &&
-                    exists $APPENDER_BY_NAME{$appendername}->{appender});
-    }
     %APPENDER_BY_NAME   = ();
     undef $INITIALIZED;
     undef $NON_INIT_WARNED;
@@ -89,7 +98,7 @@ sub reset {
 
     #clear out all the existing appenders
     foreach my $logger (values %$LOGGERS_BY_NAME){
-        $logger->{appender_names} = ();
+        $logger->{appender_names} = [];
 
 	#this next bit deals with an init_and_watch case where a category
 	#is deleted from the config file, we need to zero out the existing
@@ -97,7 +106,7 @@ sub reset {
 	#behavior --kg
         next if $logger eq $ROOT_LOGGER;
         $logger->{level} = undef;
-        $logger->level();  #set it from the heirarchy
+        $logger->level();  #set it from the hierarchy
     }
 
     # Clear all filters
@@ -140,9 +149,19 @@ sub _new {
 
    $self->set_output_methods;
 
+   print("Created logger $self ($category)\n") if _INTERNAL_DEBUG;
+
    return $self;
 }
 
+##################################################
+sub category {
+##################################################
+   my ($self) = @_;
+
+   return $self->{ category };
+}
+
 ##################################################
 sub reset_all_output_methods {
 ##################################################
@@ -206,12 +225,12 @@ sub set_output_methods {
                   if _INTERNAL_DEBUG;
             $self->{$levelname}      = $coderef;
             $self->{"is_$levelname"} = generate_is_xxx_coderef("1");
-            #$self->{"is_$levelname"} = sub { 1 };
+            print "Setting is_$levelname to 1\n" if _INTERNAL_DEBUG;
         }else{
             print "  ($priority{$levelname} > $level)\n" if _INTERNAL_DEBUG;
             $self->{$levelname}      = $noop;
             $self->{"is_$levelname"} = generate_is_xxx_coderef("0");
-            #$self->{"is_$levelname"} = sub { 0 };
+            print "Setting is_$levelname to 0\n" if _INTERNAL_DEBUG;
         }
 
         print("  Setting [$self] $self->{category}.$levelname to ",
@@ -229,25 +248,16 @@ sub generate_coderef {
     print "generate_coderef: ", scalar @$appenders, 
           " appenders\n" if _INTERNAL_DEBUG;
 
-    my $coderef = '';
-    my $watch_delay_code = '';
+    my $watch_check_code = generate_watch_code("logger", 1);
 
-    # Doing this with eval strings to sacrifice init/reload time
-    # for runtime efficiency, so the conditional won't be included
-    # if it's not needed
+    return sub {
+      my $logger = shift;
+      my $level  = pop;
 
-    if (defined $Log::Log4perl::Config::WATCHER) {
-        $watch_delay_code = generate_watch_code();
-    }
-
-    my $code = <<EOL;
-    sub {
-      my (\$logger)  = shift;
-      my (\$level)   = pop;
-      my \$message;
-      my \$appenders_fired = 0;
+      my $message;
+      my $appenders_fired = 0;
       
-      # Evaluate all parameters that need to evaluated. Two kinds:
+      # Evaluate all parameters that need to be evaluated. Two kinds:
       #
       # (1) It's a hash like { filter => "filtername",
       #                        value  => "value" }
@@ -257,76 +267,71 @@ sub generate_coderef {
       #     => coderef()
       #
 
-      \$message   = [map { ref \$_ eq "HASH" && 
-                           exists \$_->{filter} && 
-                           ref \$_->{filter} eq 'CODE' ?
-                               \$_->{filter}->(\$_->{value}) :
-                           ref \$_ eq "CODE" ?
-                               \$_->() : \$_ 
-                          } \@_];                  
-      
-      print("coderef: \$logger->{category}\n") if _INTERNAL_DEBUG;
+      $message   = [map { ref $_ eq "HASH" && 
+                           exists $_->{filter} && 
+                           ref $_->{filter} eq 'CODE' ?
+                               $_->{filter}->($_->{value}) :
+                           ref $_ eq "CODE" ?
+                               $_->() : $_ 
+                          } @_];                  
 
-      $watch_delay_code;  #note interpolation here
-      
-      foreach my \$a (\@\$appenders) {   #note the closure here
-          my (\$appender_name, \$appender) = \@\$a;
+      print("coderef: $logger->{category}\n") if _INTERNAL_DEBUG;
 
-          print("  Sending message '<\$message>' (\$level) " .
-                "to \$appender_name\n") if _INTERNAL_DEBUG;
+      if(defined $Log::Log4perl::Config::WATCHER) {
+          return unless $watch_check_code->($logger, @_, $level);
+      }
+
+      foreach my $a (@$appenders) {   #note the closure here
+          my ($appender_name, $appender) = @$a;
+
+          print("  Sending message '<$message->[0]>' ($level) " .
+                "to $appender_name\n") if _INTERNAL_DEBUG;
                 
-          \$appender->log(
+          $appender->log(
               #these get passed through to Log::Dispatch
-              { name    => \$appender_name,
-                level   => \$Log::Log4perl::Level::L4P_TO_LD{
-                               \$level},   
-                message => \$message,
+              { name    => $appender_name,
+                level   => $Log::Log4perl::Level::L4P_TO_LD{
+                               $level},   
+                message => $message,
               },
               #these we need
-              \$logger->{category},
-              \$level,
-          ) and \$appenders_fired++;
+              $logger->{category},
+              $level,
+          ) and $appenders_fired++;
               # Only counting it if it returns a true value. Otherwise
               # the appender threshold might have suppressed it after all.
     
       } #end foreach appenders
     
-      return \$appenders_fired;
+      return $appenders_fired;
 
     }; #end coderef
-
-EOL
-
-    $coderef = eval $code or die "$@";
-
-    return $coderef;
 }
 
 ##################################################
 sub generate_noop_coderef {
 ##################################################
-    my $coderef = '';
-    my $watch_delay_code = '';
+    my $watch_delay_code;
 
-    if (defined $Log::Log4perl::Config::WATCHER) {
-        $watch_delay_code = generate_watch_code();
-        $watch_delay_code = <<EOL;
-        my \$logger;
-        my \$level;
-        $watch_delay_code
-EOL
+    # This might seem crazy at first, but even in a Log4perl noop, we
+    # need to check if the configuration changed in a init_and_watch 
+    # situation. Why? Say, an application is running in a loop that
+    # constantly tries to issue debug() messages, but they're suppressed by
+    # the current Log4perl configuration. If debug() (which is a noop
+    # here) wasn't watching the configuration for changes, it would never
+    # catch the case where someone bumps up the log level and expects
+    # the application to pick it up and start logging debug() statements.
+
+    my $watch_check_code = generate_watch_code("logger", 1);
+
+    my $coderef;
+
+    if(defined $Log::Log4perl::Config::WATCHER) {
+        $coderef = $watch_check_code;
+    } else {
+        $coderef = sub { undef };
     }
 
-    my $code = <<EOL;
-    \$coderef = sub {
-        print("noop: \n") if _INTERNAL_DEBUG;
-        $watch_delay_code
-        return undef;
-     };
-EOL
-
-    eval $code or die "$@";
-
     return $coderef;
 }
 
@@ -335,62 +340,88 @@ sub generate_is_xxx_coderef {
 ##################################################
     my($return_token) = @_;
 
-    my $coderef    = sub { $return_token };
-
-    if (defined $Log::Log4perl::Config::WATCHER) {
-
-        my $cond = generate_watch_conditional();
-
-        my $watch_code = <<EOL;
-        my(\$logger, \$subname) = \@_;
-        if($cond) {
-            Log::Log4perl->init_and_watch();
-            # Forward call to new configuration
-            return \$logger->\$subname();
-        }
-EOL
-
-        my $code = <<EOL;
-        \$coderef = sub { $watch_code return $return_token; };
-EOL
-
-        eval $code or die "$@";
-    }
-
-    return $coderef;
+    return generate_watch_code("checker", $return_token);
 }
 
 ##################################################
 sub generate_watch_code {
 ##################################################
+    my($type, $return_token) = @_;
+
     print "generate_watch_code:\n" if _INTERNAL_DEBUG;
 
+      # No watcher configured, return a no-op as watch code.
+    if(! defined $Log::Log4perl::Config::WATCHER) {
+        return sub { $return_token };
+    }
+
     my $cond = generate_watch_conditional();
 
-    return <<EOL;
+    return sub {
         print "exe_watch_code:\n" if _INTERNAL_DEBUG;
-                       
-        # more closures here
-        if($cond) {
-            if(!defined \$logger) {
-                \$logger  = shift;
-                \$level   = pop;
-            }
-           
-            Log::Log4perl->init_and_watch();
-                       
-            my \$methodname = lc(\$level);
 
-                # Bump up the caller level by two, since
-                # we've artifically introduced additional levels.
-            local(\$Log::Log4perl::caller_depth);
-            \$Log::Log4perl::caller_depth += 2;
+       if(_INTERNAL_DEBUG) {
+           print "Next check: ",
+             "$Log::Log4perl::Config::Watch::NEXT_CHECK_TIME ",
+             " Now: ", time(), " Mod: ",
+             (stat($Log::Log4perl::Config::WATCHER->file()))[9],
+             "\n";
+       }
 
-            \$logger->\$methodname(\@_); # send the message
-                                         # to the new configuration
-            return;        #and return, we're done with this incarnation
-        }
-EOL
+       if( $cond->() ) {
+           my $init_permitted = 1;
+
+           if(exists $Log::Log4perl::Config::OPTS->{ preinit_callback } ) {
+               print "Calling preinit_callback\n" if _INTERNAL_DEBUG;
+               $init_permitted = 
+               $Log::Log4perl::Config::OPTS->{ preinit_callback }->( 
+                   Log::Log4perl::Config->watcher()->file() );
+               print "Callback returned $init_permitted\n" if _INTERNAL_DEBUG;
+           }
+
+           if( $init_permitted ) {
+               Log::Log4perl->init_and_watch();
+           } else {
+               # It was time to reinit, but init wasn't permitted.
+               # Return true, so that the logger continues as if
+               # it wasn't time to reinit.
+               return 1;
+           }
+
+           my $logger = shift;
+           my $level  = pop;
+
+           # Forward call to new configuration
+           if($type eq "checker") {
+               return $logger->$level();
+
+           } elsif( $type eq "logger") {
+               my $methodname = lc($level);
+
+               # Bump up the caller level by three, since
+               # we've artifically introduced additional levels.
+               local $Log::Log4perl::caller_depth =
+                     $Log::Log4perl::caller_depth + 3;
+
+               # Get a new logger for the same category (the old
+               # logger might be obsolete because of the re-init)
+               $logger = Log::Log4perl::get_logger( $logger->{category} );
+
+               $logger->$methodname(@_); # send the message
+               # to the new configuration
+               return undef;     # Return false, so the logger finishes
+               # prematurely and doesn't log the same 
+               # message again.
+           } else {
+               die "internal error: unknown type";
+           }
+       } else {
+           if(_INTERNAL_DEBUG) {
+               print "Conditional returned false\n";
+           }
+           return $return_token;
+       }
+   };
 }
 
 ##################################################
@@ -400,13 +431,16 @@ sub generate_watch_conditional {
     if(defined $Log::Log4perl::Config::Watch::SIGNAL_CAUGHT) {
         # In this mode, we just check for the variable indicating
         # that the signal has been caught
-        return q{$Log::Log4perl::Config::Watch::SIGNAL_CAUGHT};
+        return sub {
+            return $Log::Log4perl::Config::Watch::SIGNAL_CAUGHT;
+        };
     }
 
-    # In this mode, we check if the config file has been modified
-    return q{time() > $Log::Log4perl::Config::Watch::NEXT_CHECK_TIME 
-              and $Log::Log4perl::Config::WATCHER->change_detected()};
-  
+    return sub {
+        return 
+            ( time() > $Log::Log4perl::Config::Watch::NEXT_CHECK_TIME and 
+              $Log::Log4perl::Config::WATCHER->change_detected() );
+    };
 }
 
 ##################################################
@@ -529,7 +563,7 @@ sub get_logger {
     my($class, $category) = @_;
 
     unless(defined $ROOT_LOGGER) {
-        die "Internal error: Root Logger not initialized.";
+        Carp::confess "Internal error: Root Logger not initialized.";
     }
 
     return $ROOT_LOGGER if $category eq "";
@@ -543,8 +577,6 @@ sub add_appender {
 ##################################################
     my($self, $appender, $dont_reset_all) = @_;
 
-    my $not_to_dispatcher = 0;
-
         # We take this as an indicator that we're initialized.
     $INITIALIZED = 1;
 
@@ -552,18 +584,15 @@ sub add_appender {
 
     $self->{num_appenders}++;  #should this be inside the unless?
 
+      # Add newly created appender to the end of the appender array
     unless (grep{$_ eq $appender_name} @{$self->{appender_names}}){
         $self->{appender_names} = [sort @{$self->{appender_names}}, 
                                         $appender_name];
     }
 
-    if ($APPENDER_BY_NAME{$appender_name}) {
-        $not_to_dispatcher = 1;
-    }else{
-        $APPENDER_BY_NAME{$appender_name} = $appender;
-    }
+    $APPENDER_BY_NAME{$appender_name} = $appender;
 
-    &reset_all_output_methods
+    reset_all_output_methods
                 unless $dont_reset_all;  # keep us from getting overworked
                                          # if it's  the config file calling us
 
@@ -663,6 +692,7 @@ sub create_custom_level {
                            "forgot to pass in a level after which to " .
                            "place the new level!");
   my $syslog_equiv = shift; # can be undef
+  my $log_dispatch_level = shift; # optional
 
   ## only let users create custom levels before initialization
 
@@ -696,7 +726,8 @@ sub create_custom_level {
       create_custom_level("cust1", cust2);
    }) if (${Log::Log4perl::Level::LEVELS{$cust_prio}});
 
-  Log::Log4perl::Level::add_priority($level, $cust_prio, $syslog_equiv);
+  Log::Log4perl::Level::add_priority($level, $cust_prio, $syslog_equiv,
+                                     $log_dispatch_level);
 
   print("Adding prio $level at $cust_prio\n") if _INTERNAL_DEBUG;
 
@@ -740,18 +771,23 @@ sub create_log_level_methods {
   # -erik
 
   *{__PACKAGE__ . "::$lclevel"} = sub {
-        print "$lclevel: ($_[0]->{category}/$_[0]->{level}) [@_]\n" 
-            if _INTERNAL_DEBUG;
+        if(_INTERNAL_DEBUG) {
+            my $level_disp = (defined $_[0]->{level} ? $_[0]->{level} 
+                                                     : "[undef]");
+            print "$lclevel: ($_[0]->{category}/$level_disp) [@_]\n";
+        }
         init_warn() unless $INITIALIZED or $NON_INIT_WARNED;
         $_[0]->{$level}->(@_, $level) if defined $_[0]->{$level};
      };
 
   # Added these to have is_xxx functions as fast as xxx functions
   # -ms
+  
+  my $islevel   = "is_" . $level;
+  my $islclevel = "is_" . $lclevel;
 
   *{__PACKAGE__ . "::is_$lclevel"} = sub {
-      $_[0]->{"is_" . $level}->($_[0], "is_" . $lclevel) if
-          defined $_[0]->{$level};
+      $_[0]->{$islevel}->($_[0], $islclevel);
   };
   
   # Add the isXxxEnabled() methods as identical to the is_xxx
@@ -783,30 +819,24 @@ sub init_warn {
 # call me from a sub-func to spew the sub-func's caller
 #######################################################
 sub callerline {
-  # the below could all be just:
-  # my ($pack, $file, $line) = caller(2);
-  # but if we every bury this further, it'll break. So we do this
-  # little trick stolen and paraphrased from Carp/Heavy.pm
-  my($message) = @_;
+  my $message = join ('', @_);
 
-  my $i = 0;
-  my (undef, $localfile, undef) = caller($i++);
-  my ($pack, $file, $line);
-  do {
-    ($pack, $file, $line) = caller($i++);
-  } while ($file && $file eq $localfile);
+  my $caller_offset = 
+    Log::Log4perl::caller_depth_offset( 
+        $Log::Log4perl::caller_depth + 1 );
 
-  my $has_newline;
+  my ($pack, $file, $line) = caller($caller_offset);
 
-  $has_newline++ if chomp $message;
+  if (not chomp $message) {     # no newline
+    $message .= " at $file line $line";
 
-  $message .= " at $file line $line" if !$has_newline;
-
-  # Someday, we'll use Threads. Really.
-  if (defined &Thread::tid) {
-    my $tid = Thread->self->tid;
-    $message .= " thread $tid" if $tid and !$has_newline;
+    # Someday, we'll use Threads. Really.
+    if (defined &Thread::tid) {
+      my $tid = Thread->self->tid;
+      $message .= " thread $tid" if $tid;
+    }
   }
+
   return ($message, "\n");
 }
 
@@ -814,44 +844,58 @@ sub callerline {
 sub and_warn {
 #######################################################
   my $self = shift;
-  my $msg = join("", @_[0 .. $#_]);
-  CORE::warn(callerline($msg));
+  CORE::warn(callerline($self->warning_render(@_)));
 }
 
 #######################################################
 sub and_die {
 #######################################################
   my $self = shift;
-  my $msg = join("", @_[0 .. $#_]);
-  die(callerline($msg));
+  my $arg  = $_[0];
+
+  my($msg) = callerline($self->warning_render(@_));
+
+  if($DIE_DEBUG) {
+      $DIE_DEBUG_BUFFER = "DIE_DEBUG: $msg";
+  } else {
+      if( $Log::Log4perl::STRINGIFY_DIE_MESSAGE ) {
+          die("$msg\n");
+      }
+      die $arg;
+  }
 }
 
 ##################################################
 sub logwarn {
 ##################################################
   my $self = shift;
+
+  local $Log::Log4perl::caller_depth = 
+        $Log::Log4perl::caller_depth + 1;
+
   if ($self->is_warn()) {
         # Since we're one caller level off now, compensate for that.
-    $Log::Log4perl::caller_depth++;
     my @chomped = @_;
     chomp($chomped[-1]);
     $self->warn(@chomped);
-    $Log::Log4perl::caller_depth--;
-    $self->and_warn(@_);
   }
+
+  $self->and_warn(@_);
 }
 
 ##################################################
 sub logdie {
 ##################################################
   my $self = shift;
+
+  local $Log::Log4perl::caller_depth = 
+        $Log::Log4perl::caller_depth + 1;
+
   if ($self->is_fatal()) {
         # Since we're one caller level off now, compensate for that.
-    $Log::Log4perl::caller_depth++;
     my @chomped = @_;
     chomp($chomped[-1]);
     $self->fatal(@chomped);
-    $Log::Log4perl::caller_depth--;
   }
 
   $Log::Log4perl::LOGDIE_MESSAGE_ON_STDERR ? 
@@ -864,41 +908,41 @@ sub logexit {
 ##################################################
   my $self = shift;
 
+  local $Log::Log4perl::caller_depth = 
+        $Log::Log4perl::caller_depth + 1;
+
   if ($self->is_fatal()) {
         # Since we're one caller level off now, compensate for that.
-    $Log::Log4perl::caller_depth++;
     my @chomped = @_;
     chomp($chomped[-1]);
     $self->fatal(@chomped);
-    $Log::Log4perl::caller_depth--;
   }
 
   exit $Log::Log4perl::LOGEXIT_CODE;
 }
 
-##################################################
-# for die and warn, carp long/shortmess return line #s and the like
-##################################################
-sub noop {
-##################################################
-  return @_;
-}
-
 ##################################################
 # clucks and carps are WARN level
 sub logcluck {
 ##################################################
   my $self = shift;
 
-  local $Carp::CarpLevel = $Carp::CarpLevel + 1;
+  local $Log::Log4perl::caller_depth = 
+        $Log::Log4perl::caller_depth + 1;
+
+  local $Carp::CarpLevel = 
+        $Carp::CarpLevel + 1;
+
+  my $msg = $self->warning_render(@_);
 
   if ($self->is_warn()) {
-    my $message = Carp::longmess(@_);
+    my $message = Carp::longmess($msg);
     foreach (split(/\n/, $message)) {
       $self->warn("$_\n");
     }
-    Carp::cluck(noop($message));
   }
+
+  Carp::cluck($msg);
 }
 
 ##################################################
@@ -908,13 +952,19 @@ sub logcarp {
 
   local $Carp::CarpLevel = $Carp::CarpLevel + 1;
 
+  local $Log::Log4perl::caller_depth = 
+        $Log::Log4perl::caller_depth + 1;
+
+  my $msg = $self->warning_render(@_);
+
   if ($self->is_warn()) {
-    my $message = Carp::shortmess(@_);
+    my $message = Carp::shortmess($msg);
     foreach (split(/\n/, $message)) {
       $self->warn("$_\n");
     }
-    Carp::carp(noop($message)) if $Log::Log4perl::LOGDIE_MESSAGE_ON_STDERR;
   }
+
+  Carp::carp($msg);
 }
 
 ##################################################
@@ -923,18 +973,31 @@ sub logcarp {
 sub logcroak {
 ##################################################
   my $self = shift;
+  my $arg  = $_[0];
 
-  local $Carp::CarpLevel = $Carp::CarpLevel + 1;
+  my $msg = $self->warning_render(@_);
+
+  local $Carp::CarpLevel = 
+        $Carp::CarpLevel + 1;
+
+  local $Log::Log4perl::caller_depth = 
+        $Log::Log4perl::caller_depth + 1;
 
-  my $message = Carp::shortmess(@_);
   if ($self->is_fatal()) {
+    my $message = Carp::shortmess($msg);
     foreach (split(/\n/, $message)) {
       $self->fatal("$_\n");
     }
   }
 
+  my $croak_msg = $arg;
+
+  if( $Log::Log4perl::STRINGIFY_DIE_MESSAGE ) {
+      $croak_msg = $msg;
+  }
+
   $Log::Log4perl::LOGDIE_MESSAGE_ON_STDERR ? 
-      Carp::croak(noop($message)) : 
+      Carp::croak($croak_msg) : 
         exit($Log::Log4perl::LOGEXIT_CODE);
 }
 
@@ -942,18 +1005,31 @@ sub logcroak {
 sub logconfess {
 ##################################################
   my $self = shift;
+  my $arg  = $_[0];
 
-#local $Carp::CarpLevel = 2;
+  local $Carp::CarpLevel = 
+        $Carp::CarpLevel + 1;
+
+  local $Log::Log4perl::caller_depth = 
+        $Log::Log4perl::caller_depth + 1;
+
+  my $msg = $self->warning_render(@_);
 
-  my $message = Carp::longmess(@_);
   if ($self->is_fatal()) {
+    my $message = Carp::longmess($msg);
     foreach (split(/\n/, $message)) {
       $self->fatal("$_\n");
     }
   }
 
+  my $confess_msg = $arg;
+
+  if( $Log::Log4perl::STRINGIFY_DIE_MESSAGE ) {
+      $confess_msg = $msg;
+  }
+
   $Log::Log4perl::LOGDIE_MESSAGE_ON_STDERR ? 
-      die(noop($message)) :
+      confess($confess_msg) :
         exit($Log::Log4perl::LOGEXIT_CODE);
 }
 
@@ -963,24 +1039,33 @@ sub logconfess {
 sub error_warn {
 ##################################################
   my $self = shift;
+
+  local $Log::Log4perl::caller_depth = 
+        $Log::Log4perl::caller_depth + 1;
+
   if ($self->is_error()) {
     $self->error(@_);
-    $self->and_warn(@_);
   }
+
+  $self->and_warn(@_);
 }
 
 ##################################################
 sub error_die {
 ##################################################
   my $self = shift;
+
+  local $Log::Log4perl::caller_depth = 
+        $Log::Log4perl::caller_depth + 1;
+
+  my $msg = $self->warning_render(@_);
+
   if ($self->is_error()) {
-    $Log::Log4perl::caller_depth++;
-    $self->error(@_);
-    $Log::Log4perl::caller_depth--;
+    $self->error($msg);
   }
 
   $Log::Log4perl::LOGDIE_MESSAGE_ON_STDERR ? 
-      $self->and_die(@_) :
+      $self->and_die($msg) :
         exit($Log::Log4perl::LOGEXIT_CODE);
 }
 
@@ -1023,8 +1108,6 @@ sub dec_level {
     $self->set_output_methods;
 }
 
-##################################################
-
 1;
 
 __END__
@@ -1042,11 +1125,34 @@ Log::Log4perl::Logger - Main Logger Class
 While everything that makes Log4perl tick is implemented here,
 please refer to L<Log::Log4perl> for documentation.
 
-=head1 SEE ALSO
+=head1 LICENSE
+
+Copyright 2002-2012 by Mike Schilli E<lt>m@perlmeister.comE<gt> 
+and Kevin Goess E<lt>cpan@goess.orgE<gt>.
+
+This library is free software; you can redistribute it and/or modify
+it under the same terms as Perl itself. 
 
 =head1 AUTHOR
 
-    Mike Schilli, <log4perl@perlmeister.com>
-    Kevin Goess, <cpan@goess.org>
+Please contribute patches to the project on Github:
+
+    http://github.com/mschilli/log4perl
+
+Send bug reports or requests for enhancements to the authors via our
+
+MAILING LIST (questions, bug reports, suggestions/patches): 
+log4perl-devel@lists.sourceforge.net
+
+Authors (please contact them via the list above, not directly):
+Mike Schilli <m@perlmeister.com>,
+Kevin Goess <cpan@goess.org>
+
+Contributors (in alphabetical order):
+Ateeq Altaf, Cory Bennett, Jens Berthold, Jeremy Bopp, Hutton
+Davidson, Chris R. Donnelly, Matisse Enzer, Hugh Esco, Anthony
+Foiani, James FitzGibbon, Carl Franks, Dennis Gregorovic, Andy
+Grundman, Paul Harrington, David Hull, Robert Jacobson, Jason Kohles, 
+Jeff Macdonald, Markus Peter, Brett Rann, Peter Rabbitson, Erik
+Selberg, Aaron Straup Cope, Lars Thegler, David Viner, Mac Yang.
 
-=cut
